r"""
Factored denominators
"""
#*****************************************************************************
#       Copyright (C) 2019 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
#*****************************************************************************

from __future__ import absolute_import, print_function

from sage.rings.integer_ring import ZZ
from sage.modules.vector_integer_dense import Vector_integer_dense
from sage.structure.element import parent

def vector_to_monomial_string(u, var_names):
    r"""
    EXAMPLES::

        sage: from surface_dynamics.misc.factored_denominator import vector_to_monomial_string
        sage: vector_to_monomial_string((0,4,3), 'x')
        'x1^4*x2^3'
        sage: vector_to_monomial_string((0,4,3), 'hello')
        'hello1^4*hello2^3'
        sage: vector_to_monomial_string((0,4,3), ('x','y','z'))
        'y^4*z^3'
        sage: vector_to_monomial_string((1,0,-1), 'x')
        'x0*x2^-1'
    """
    s = []
    for i,j in enumerate(u):
        if j:
            if isinstance(var_names, str):
                var = '%s%d' % (var_names, i)
            else:
                var = var_names[i]

            if j == 1:
                s.append(var)
            else:
                s.append('%s^%d' % (var, j))

    return '*'.join(s) if s else '1'

def vector_to_linear_form_string(u, var_names):
    r"""
    EXAMPLES::

        sage: from surface_dynamics.misc.factored_denominator import vector_to_linear_form_string
        sage: vector_to_linear_form_string((0,4,3), 'x')
        '4*x1 + 3*x2'
        sage: vector_to_linear_form_string((0,4,3), 'hello')
        '4*hello1 + 3*hello2'
        sage: vector_to_linear_form_string((0,4,3), ('x','y','z'))
        '4*y + 3*z'
        sage: vector_to_linear_form_string((1,0,-1), 'x')
        'x0 + -x2'
    """
    s = []
    for i,j in enumerate(u):
        if j:
            if isinstance(var_names, str):
                var = '%s%d' % (var_names, i)
            else:
                var = var_names[i]

            if j == 1:
                s.append(var)
            elif j == -1:
                s.append('-' + var)
            else:
                s.append('%d*%s' % (j, var))

    return ' + '.join(s) if s else '0'



# NOTE: should this be an instance of Factorization?
class FactoredDenominator(object):
    r"""
    Factored denominator

    This class is a simple datastructure to handle factored denominator, that is
    a list of pairs ``(v, m)`` where the ``v`` are integer vectors (of fixed
    dimension) and ``m`` are multiplicities (i.e. positive integers).

    It is used for at least two purposes:
    
    - (Factored) product of polynomials of the form `(1 - m)^d` where `m` is a
      monomial

    - generalized multiple zeta values, where denominators are products of
      linear forms

    EXAMPLES::

        sage: from surface_dynamics.misc.factored_denominator import FactoredDenominator

        sage: V = ZZ**3
        sage: f1 = FactoredDenominator([((1,0,0), 2)], V)
        sage: f2 = FactoredDenominator([((0,1,2), 3), ((1,1,1), 1)], V)
        sage: f3 = FactoredDenominator([((0,-1,2), 1), ((1,0,0), 1), ((0,0,2), 1)], V)
        sage: f1
        {(1, 0, 0): 2}
        sage: f1 * f2 * f3
        {(1, 0, 0): 3, (0, 1, 2): 3, (1, 1, 1): 1, (0, -1, 2): 1, (0, 0, 2): 1}
        sage: hash(f1)  # random
        9164823176457928306

        sage: FactoredDenominator(f1)
        {(1, 0, 0): 2}
    """
    __slots__ = ['_dict', '_tuple']

    def __init__(self, data, V=None):
        if isinstance(data, (tuple, list)):
            if V is None:
                raise ValueError("a ZZ-free module V must be provided")
            d = V.dimension()
            self._dict = {}
            for mon, mult in data:
                mon = V(mon)
                if mon.is_zero():
                    raise ValueError('zero in denominator')
                for i in range(d):
                    if mon[i]:
                        break
                mult = ZZ(mult)
                if mult <= 0:
                    raise ValueError('non-positive multiplicity in denominator')
                mon.set_immutable()
                self._dict[mon] = mult

        elif isinstance(data, dict):
            if V is not None:
                self._dict = {}
                for k,v in data.items():
                    k = V(k)
                    k.set_immutable()
                    self._dict[k] = v
            else:
                self._dict = data

        elif isinstance(data, FactoredDenominator):
            self._dict = data._dict.copy()
            self._tuple = data._tuple[:]
            return

        elif isinstance(data, Vector_integer_dense):
            data = V(data)
            data.set_immutable()
            self._dict = {data: ZZ.one()}
            self._tuple = ((data, ZZ.one()),)
            return

        else:
            raise TypeError('invalid data of type {} to initialized a FactoredDenominator'.format(type(data)))

        self._tuple = tuple(sorted(self._dict.items()))

    def __len__(self):
        r"""
        Return the number of factors (without multiplicities).
        """
        return len(self._tuple)

    def __iter__(self):
        r"""
        Iterates through the pairs ``(monomial, exponent)``.
        """
        return iter(self._tuple)

    def degree(self):
        r"""
        EXAMPLES::

            sage: from surface_dynamics.misc.factored_denominator import FactoredDenominator

            sage: V = ZZ**3
            sage: FactoredDenominator([((1,0,0), 2)], V).degree()
            2
            sage: FactoredDenominator([((0,1,2), 3), ((1,1,1), 1)], V).degree()
            4
            sage: FactoredDenominator([((0,-1,2), 1), ((1,0,0), 1), ((0,0,2), 1)], V).degree()
            3
        """
        return sum(x[1] for x in self._tuple)

    def to_polynomial(self, S, extra_var=False):
        r"""
        Return the product of the term in a given polynomial ring ``S``.

        EXAMPLES::

            sage: from surface_dynamics.misc.factored_denominator import FactoredDenominator

            sage: V = ZZ ** 3
            sage: f = FactoredDenominator([((1,0,0), 2)], V)
            sage: g = FactoredDenominator([((1,0,0), 2), ((1,1,1),1)], V)

            sage: R1 = QQ['x,y,z']
            sage: f.to_polynomial(R1)
            x^2 - 2*x + 1
            sage: g.to_polynomial(R1)
            -x^3*y*z + 2*x^2*y*z - x*y*z + x^2 - 2*x + 1

            sage: f.to_polynomial(R1['E'], extra_var=True)
            x^2*E^2 - 2*x*E + 1
            sage: g.to_polynomial(R1['E'], extra_var=True)
            -x^3*y*z*E^5 + 2*x^2*y*z*E^4 - x*y*z*E^3 + x^2*E^2 - 2*x*E + 1

            sage: g.to_polynomial(QQ['t1,t2,t3'])
            -t1^3*t2*t3 + 2*t1^2*t2*t3 - t1*t2*t3 + t1^2 - 2*t1 + 1
        """
        if extra_var:
            R = S.base_ring()
            T = S.gen()
        else:
            R = S
            T = R.one()

        ans = S.one()
        for a,i in self._tuple:
            ans *= (S.one() - R.monomial(*a) * T ** sum(a)) ** i
        return ans

    def inverse_series_trunc(self, S, prec):
        r"""
        Return a truncation of the inverse series of this element.

        INPUT:

        - ``S`` - polynomial ring in one variable whose base ring is a
          multivariate polynomial ring with the same number of variables
          as this factor.
        
        - ``prec`` - precision of the Taylor expansion (in terms of total degree)

        EXAMPLES::

            sage: from surface_dynamics.misc.factored_denominator import FactoredDenominator

            sage: V = ZZ**2
            sage: R = QQ['t,u']
            sage: f = FactoredDenominator([((2,3), 2), ((4,1), 1), ((1,1), 1)], V)
            sage: f.inverse_series_trunc(R['X'], 10)
            (t^6*u^3 + 2*t^4*u^5)*X^9 +
            t^4*u^4*X^8 +
            (t^5*u^2 + 2*t^3*u^4)*X^7 +
            t^3*u^3*X^6 +
            (t^4*u + 2*t^2*u^3)*X^5 +
            t^2*u^2*X^4 +
            t*u*X^2 +
            1
        """
        # this is a very stupid way of doing so
        # we should Taylor expand each monomial and then use _mul_trunc_
        return self.to_polynomial(S, extra_var=True).inverse_series_trunc(prec)

    def degree(self):
        r"""
        EXAMPLES::

            sage: from surface_dynamics.misc.factored_denominator import FactoredDenominator

            sage: V = ZZ**3
            sage: FactoredDenominator([], V).degree()
            0
            sage: FactoredDenominator([((1,0,0), 2)], V).degree()
            2
            sage: FactoredDenominator([((0,1,2), 3), ((1,1,1), 1)], V).degree()
            4
            sage: FactoredDenominator([((0,-1,2), 1), ((1,0,0), 1), ((0,0,2), 1)], V).degree()
            3

        TESTS::

            sage: parent(FactoredDenominator([], V).degree())
            Integer Ring
            sage: parent(FactoredDenominator([((1,0,0), 2)], V).degree())
            Integer Ring
        """
        if not self._tuple:
            return ZZ.zero()
        return sum(m for _,m in self._tuple)

    # TODO
    # this method does not really make sense at this level of generality!
    # this can be defined on denominators and extended by linearity
    # we should just provide a method to do this in an abstract parent algebra
    #
    # _image_from_image_of_denominators(function, elt)
    #   return sum(num * function(den) for num,den in elt)
    def logarithmic_minus_derivative(self, j):
        r"""
        Minus the logarithmic derivative -v' / v taken with respect to the ``j``-th variable

        Since this denominator has the form 1 / (1 - m1) (1 - m2) ... (1 - mk) the
        logarithmic derivative is simply

        m1'/(1-m1) + m2'/(1-m2) + ... + mk'/(1-mk)

        INPUT:

        - ``j`` -- index of a variable

        OUTPUT: itertor of triples ``(mult, v, monomial)``

        EXAMPLES::

            sage: from surface_dynamics.misc.factored_denominator import FactoredDenominator

            sage: V = ZZ**3

            sage: f = FactoredDenominator([((1,0,0), 2)], V)
            sage: f
            {(1, 0, 0): 2}
            sage: list(f.logarithmic_minus_derivative(0))
            [(2, (0, 0, 0), (1, 0, 0))]
            sage: list(f.logarithmic_minus_derivative(1))
            []
            sage: list(f.logarithmic_minus_derivative(2))
            []

            sage: f = FactoredDenominator([((1,1,1), 1)], V)
            sage: list(f.logarithmic_minus_derivative(0))
            [(1, (0, 1, 1), (1, 1, 1))]

            sage: f = FactoredDenominator([((1,0,0), 1), ((0,1,0), 1), ((0,0,1), 1)], V)
            sage: list(f.logarithmic_minus_derivative(0))
            [(1, (0, 0, 0), (1, 0, 0))]

            sage: f = FactoredDenominator([((1,0,0), 2), ((1,1,0), 3), ((1,1,1), 1)], V)
            sage: f
            {(1, 0, 0): 2, (1, 1, 0): 3, (1, 1, 1): 1}
            sage: list(f.logarithmic_minus_derivative(0))
            [(2, (0, 0, 0), (1, 0, 0)),
             (3, (0, 1, 0), (1, 1, 0)),
             (1, (0, 1, 1), (1, 1, 1))]
        """
        for mon, mult in self._tuple:
            if mon[j]:
                v = mon.__copy__()
                mult *= v[j]
                v[j] -= 1
                yield (mult, v, mon)

    def __eq__(self, other):
        r"""
        Equality test

        EXAMPLES::

            sage: from surface_dynamics.misc.factored_denominator import FactoredDenominator

            sage: V = ZZ**3

            sage: f1 = FactoredDenominator([((1,0,0), 2)], V)
            sage: f2 = FactoredDenominator([((1,0,0), 2)], V)
            sage: f3 = FactoredDenominator([((1,0,0), 1)], V)
            sage: f4 = FactoredDenominator([((1,0,1), 2)], V)

            sage: f1 == f1 and f1 == f2
            True
            sage: f1 == f3 or f1 == f4
            False
        """
        if type(self) is not type(other):
            raise TypeError
        return self._tuple == other._tuple

    def __ne__(self, other):
        if type(self) is not type(other):
            raise TypeError
        return self._tuple != other._tuple

    def __lt__(self, other):
        if type(self) is not type(other):
            raise TypeError
        return self._tuple < other._tuple

    def __le__(self, other):
        if type(self) is not type(other):
            raise TypeError
        return self._tuple <= other._tuple

    def __gt__(self, other):
        if type(self) is not type(other):
            raise TypeError
        return self._tuple > other._tuple

    def __ge__(self, other):
        if type(self) is not type(other):
            raise TypeError
        return self._tuple >= other._tuple

    def copy(self):
        res = FactoredDenominator.__new__(FactoredDenominator)
        res._dict = self._dict.copy()
        res._tuple = self._tuple
        return res

    def __mul__(self, other):
        r"""
        Multiplication

        TESTS::

            sage: from surface_dynamics.misc.factored_denominator import FactoredDenominator
            sage: V = ZZ**3

            sage: f = FactoredDenominator([((1,0,0), 2)], V)
        """
        if type(self) is not type(other):
            raise TypeError

        new_data = self._dict.copy()
        for i,j in other._dict.items():
            if i in new_data:
                new_data[i] += j
                if new_data[i].is_zero():
                    del new_data[i]
            else:
                new_data[i] = j

        return FactoredDenominator(new_data, None)

    def __truediv__(self, other):
        r"""
        (Partial) division.

        EXAMPLES::

            sage: from surface_dynamics.misc.factored_denominator import FactoredDenominator
            sage: V = ZZ**3

            sage: f1 = FactoredDenominator([((1,0,0), 3)], V)
            sage: f2 = FactoredDenominator([((1,0,0), 1)], V)
            sage: f1/f2
            {(1, 0, 0): 2}
        """
        if type(self) != type(other):
            raise TypeError

        sd = self._dict
        od = other._dict
        nd = sd.copy()
        for i,j in od.items():
            jj = sd.get(i, -1)
            if j > jj:
                raise ArithmeticError
            elif j == jj:
                del nd[i]
            else:
                nd[i] -= j

        return FactoredDenominator(nd, None)

    __div__ = __truediv__

    def __nonzero__(self):
        return bool(self._dict)

    def is_one(self):
        return not self._dict

    def str(self):
        return str(self._dict)

    def __repr__(self):
        return repr(self._dict)

    def __hash__(self):
        return hash(self._tuple)

    def lcm_update(self, other):
        if type(self) is not type(other):
            raise TypeError

        sd = self._dict
        od = other._dict
        for i,j in od.items():
            if j > sd.get(i, -1):
                sd[i] = j

        self._tuple = tuple(sorted(self._dict.items()))

    def lcm(self, other):
        r"""
        Return the lcm of two factored denomiator.

        EXAMPLES::

            sage: from surface_dynamics.misc.factored_denominator import FactoredDenominator

            sage: V = ZZ**3

            sage: f1 = FactoredDenominator([((1,0,0), 2),((0,1,0),1)], V)
            sage: f2 = FactoredDenominator([((1,0,0), 1),((0,1,0),2)], V)
            sage: f3 = FactoredDenominator([((1,0,0), 1),((0,1,0),1),((0,0,1),1)], V)
            sage: f4 = FactoredDenominator([], V)

            sage: f1.lcm(f2)
            {(1, 0, 0): 2, (0, 1, 0): 2}
            sage: f1.lcm(f3)
            {(1, 0, 0): 2, (0, 1, 0): 1, (0, 0, 1): 1}
            sage: f2.lcm(f3)
            {(1, 0, 0): 1, (0, 1, 0): 2, (0, 0, 1): 1}

            sage: f1.lcm(f4) == f1 and f4.lcm(f1) == f1
            True
        """
        res = self.copy()
        res.lcm_update(other)
        return res

    def gcd_update(self, other):
        if type(self) is not type(other):
            raise TypeError

        sd = self._dict
        od = other._dict
        for i,j in od.items():
            if j < sd.get(i, j):
                sd[i] = j

        self._tuple = tuple(sorted(self._dict.items()))

    def gcd(self, other):
        r"""
        Return the gcd of two factored denominator.

        EXAMPLES::

            sage: from surface_dynamics.misc.factored_denominator import FactoredDenominator

            sage: V = ZZ**3

            sage: f1 = FactoredDenominator([((1,0,0), 2),((0,1,0),1)], V)
            sage: f2 = FactoredDenominator([((1,0,0), 1),((0,1,0),2)], V)
            sage: f3 = FactoredDenominator([((1,0,0), 1),((0,1,0),1),((0,0,1),1)], V)
            sage: f4 = FactoredDenominator([], V)

            sage: f1.gcd(f2)
            {(1, 0, 0): 1, (0, 1, 0): 1}
            sage: f1.gcd(f2) == f1.gcd(f3) == f2.gcd(f3)
            True

            sage: f4.gcd(f1) == f4 and f1.gcd(f4) == f1
            True
        """
        res = self.copy()
        res.gcd_update(other)
        return res

    # TODO
    # this does not make much sense here
    def str_monomials(self, var_names='x'):
        terms = []
        for mon, mul in self._tuple:
            term = '(1 - %s)' % vector_to_monomial_string(mon, var_names)
            if not mul.is_one():
                term += '^%d' % mul
            terms.append(term)
        return '*'.join(terms)

    # TODO
    # this does not make much sense here
    def str_linear(self, var_names='h'):
        terms = []
        for mon, mul in self._tuple:
            term = '(' + vector_to_linear_form_string(mon, var_names) + ')'
            if not mul.is_one():
                term += '^%d' % mul
            terms.append(term)

        return '*'.join(terms)




r"""
Rational multivariate quasi-polynomial functions

Given a rational function `F` in the variables `x1, \ldots, xm` with no
pole at zero can be expanded as a (Laurent) power series `\sum f(i0, i1, ...,
im) x0^{i0} ... xm^{im}`. This modules deal with the case when `f` are quasi-polynomial
are equivalently when `F` is a finite sum of rational functions of the form

.. MATH::

    \frac{P}{\prod (1 - m_i)^{d_i}}

where the `m_i` are monomials and `d_i` are positive integers.

REFERENCES:

DH:03

[LHTY] De Loera, R. Hemmecke, J. Tauzer, R. Yoshida
 "Effective lattice point counting in rational convex polytopes"
 (LattE)

[Ba] Barvinok

[St] Stanley
"Enumerative Combinatorics"
Volume I
Cambridge Studies in Advanced Mathematics 49 (1997)

.. TODO::

 - make it barvinok compatible, the data-structure is
   a term is a ``struct short_rat`` while the generating
   series is a ``struct gen_fun``
"""
#*****************************************************************************
#       Copyright (C) 2019 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
#*****************************************************************************

from __future__ import print_function, absolute_import
from six.moves import range, map, filter, zip
from six import iteritems

from .factored_denominator import FactoredDenominator

import numbers

from sage.misc.cachefunc import cached_method

from sage.structure.sage_object import SageObject
from sage.structure.parent import Parent
from sage.structure.element import Element, parent

try:
    from sage.structure.richcmp import op_LT, op_EQ, op_NE, op_LE, op_GE, op_GT, op_LT
except ImportError:
    # sage version <= 7.6
    from sage.structure.sage_object import op_LT, op_EQ, op_NE, op_LE, op_GE, op_GT, op_LT


from sage.categories.all import CommutativeAlgebras, Rings

from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ
from sage.modules.vector_integer_dense import Vector_integer_dense



def laurent_monomial(R, arg):
    r"""
    EXAMPLES::

        sage: from surface_dynamics.misc.multivariate_generating_series import laurent_monomial
        sage: R = LaurentPolynomialRing(QQ, 'x', 2)
        sage: laurent_monomial(R, (1,2))
        x0*x1^2
        sage: laurent_monomial(R, vector((-1r,3r)))
        x0^-1*x1^3
        sage: laurent_monomial(R, (-1,3)) * laurent_monomial(R, (1,1)) == laurent_monomial(R, (0,4))
        True
    """
    # only in beta 8.1 versions of Sage
    # return R.monomial(*arg)
    arg = tuple(arg)
    if len(arg) != R.ngens():
        raise TypeError("tuple key must have same length as ngens")

    from sage.misc.misc_c import prod
    return prod(x**int(i) for (x,i) in zip(R.gens(), arg))

# custom latte count
def latte_generating_series(L, M=None):
    r"""
    EXAMPLES::

        sage: from surface_dynamics.misc.multivariate_generating_series import latte_generating_series
        sage: ieqs = [[0, 1, 0, 0, 0, 0, 0],
        ....:         [0, 0, 1, -1, 1, 0, 0],
        ....:         [0, 0, 0, 0, 1, 0, 0],
        ....:         [0, 0, 0, 1, 0, 0, 0],
        ....:         [0, 0, 1, 0, 0, 0, 0]]
        sage: eqns = [[0, 0, 1, -1, 1, 0, -1], [0, 0, 1, -1, 1, -1, 0]]
        sage: L = Polyhedron(ieqs=ieqs, eqns=eqns)
        sage: latte_generating_series(L)   # optional - latte_int
        (1)/((1 - x2^-1*x4*x5)*(1 - x2*x3)*(1 - x1*x2)*(1 - x0)) + (1)/((1 - x3*x4*x5)*(1 - x2*x4^-1*x5^-1)*(1 - x1*x4*x5)*(1 - x0))
    """
    if M is None:
        M = MultivariateGeneratingSeriesRing('x', L.ambient_dim())
    try:
        from sage.interfaces.latte import count
    except ImportError:
        from sage.version import version
        raise ValueError('your Sage version is too old ({}) to use this function'.format(version))
    ans = count(L.cdd_Hrepresentation(), cdd=True, multivariate_generating_function=True, raw_output=True)
    return parse_latte_generating_series(M, ans)

def parse_latte_generating_series(M, s):
    r"""
    INPUT:

    - ``M`` - a multivariate generating series ring

    - ``s`` - a string as given by LattE

    OUTPUT: multivariate short rational function
    """
    n = M.ngens()
    R = M.laurent_polynomial_ring()
    V = M.free_module()

    terms = s.strip().split('\n + ')
    m = M.zero()
    for term in terms:
        num, den = term.split('/')
        m_den = {}
        for term in den[2:-2].split(')*('):
            mon = V(0)
            assert term.startswith('1-')
            term = term[2:].split('*')
            for v in term:
                if '^' in v:
                    v, mult = v.split('^')
                    if mult.startswith('(') and mult.endswith(')'):
                        mult = mult[1:-1]
                    mult = int(mult)
                else:
                    mult = 1

                assert v.startswith('x[') and v.endswith(']'), (s, term, v)
                i = int(v[2:-1])
                mon[i] += mult
            mon.set_immutable()
            if mon in m_den:
                m_den[mon] += 1
            else:
                m_den[mon] = 1
        m += M.term(R(num), list(m_den.items()))

    return m


def monomial_projection(num, den, i, M):
    r"""
    Return the monomial substitution corresponding to setting the ``i``-th variable
    to ``1`` in the term ``num / den`` of the multivariate series ring ``M``.

    INPUT:

    - ``num`` - multivariate polynomial

    - ``den`` - a factored denominator

    - ``i`` - (integer) the index of a variable

    - ``M`` - multivariate generating series ring


    EXAMPLES::

        sage:
    """
    R = parent(num)
    n = R.ngens()
    x = R.gen(i)

    # formally we replace by (1+t) and let t tend to 0
    B = []   # singular terms ("bad")
    G = []   # regular terms  ("good")
    for v, mult in den:
        if v[i] != 0 and sum(v[j] != 0 for j in range(n)) == 1:
            B.append((v[i], mult))
        else:
            G.append((v,mult))


    # 1. regular case
    if not B:
        new_dict = {}
        for v,mult in G:
            v = v.__copy__()
            v[i] = 0
            v.set_immutable()
            if v in new_dict:
                new_dict[v] += mult
            else:
                new_dict[v] = mult

        num = num.subs({x: 1})
        return M.term(num, den)

    # 2. singular case
    r = sum(x[1] for x in B) + 1   # r-1 is the degree of the pole
    S = M.laurent_polynomial_ring_extra_var()
    SS = M.with_extra_var()
    t = S.gen()
    tt = SS.gen()

    P = num.subs({x: 1+t}).truncate(r)

    # 2.a) each singular term is 1 / (1 - (1+t)^j)^mult
    for j,mult in B:
        q = ((1 - (1 + t)**j) // t).inverse_series_trunc(r)._power_trunc(mult, r)
        P = q._mul_trunc_(P, r)

    # 2.b) each regular term is 1 / (1 - (1+t)^j s)^mult
    for v,mult in G:
        if v[i] == 0:  # no t involved
            P *= M.term(1, [(v,mult)])
        else:          # develop in powers of t
            j = v[i]
            v = v.__copy__()
            v[i] = 0
            v.set_immutable()
            s = laurent_monomial(S, v)
            z = M.term(1, [(1-s), 1])
            q = (z * (1 - ((1+t)**j - 1) * s * z).inverse_series_trunc(r))._power_trunc(mult, r)
            P = q._mul_trunc_(P, r)

    return P[r-1]



#def monomial_substitution(mat, num, den, M):
#    r"""
#    Apply the monomial substitution on ``num`` / ``den`` given by ``mat``.
#
#    We assume that the matrix applies on rows (in particular it should
#    have the same number of rows as the dimension of the monomial in
#    this factored denominator).
#
#    If some monomial is in the kernel, a vector ``c`` needs to be provided
#    and in that case this function return the constant term of the expansion
#    in the direction given by ``c`` (the result might depend on the choice of
#    the vector ``c`` provided).
#
#    INPUT:
#
#    - ``num`` - a polynomial
#
#    - ``den`` - a factored denominator
#
#    - ``M`` - a multivariate series ring
#
#    - ``c`` - an optional vector
#
#    EXAMPLES::
#
#        sage: from surface_dynamics.misc.multivariate_generating_series import FactoredDenominator
#        sage: V = ZZ**3
#        sage: f = FactoredDenominator([((1,0,0), 1),((0,1,0),1),((0,0,1),1)], V)
#        sage: f.monomial_substitution(matrix(3, 2, [1,1,1,0,1,0]))
#        (1 - x0)^2*(1 - x0*x1)
#        sage: f.monomial_substitution(matrix(3, 4, [1,1,1,1,1,1,0,0,0,0,1,1]))
#        (1 - x2*x3)*(1 - x0*x1)*(1 - x0*x1*x2*x3)
#    """
#    B = []
#    G = []
#    for a,i in self._tuple:
#        b = a * mat
#        if b.is_zero():
#            B.append((a,i))
#        else:
#            b.set_immutable()
#            G.append((a,b,i))
#
#    # no singular part
#    if not B:
#        new_dict = {}
#        for a,b,i in G:
#            if b in new_dict:
#                new_dict[b] += i
#            else:
#                new_dict[b] = i
#        return M.term(1, new_dict)
#
#    # case with singular part
#    l = sum(i for a,i in B)  # degree of the singular part
#    # XXX
#    raise NotImplementedError('singular case not implemented')




# This should be a real number!
# ... and we need an algebra to handle all these values properly
class _OLD_TOTO_(object):
    r"""
    A generalized multiple zeta value.

    A *generalized multiple zeta value* is a sum of the form

    .. MATH::

        \sum_{h in ZZ^s_+} chi(h) /  \ell_1(h)^{-1} \ell_2(h)^{-1} ... \ell_d(h)^{-1}

    where `chi(h)` is a character, `\ell_1(h)`, `\ell_2(h)`, ..., `\ell_d(h)`
    are linear form with non-negative coefficients. The integer `d` is called
    the *weight*.

    The version with character is not implemented!
    """
    def gp_multizeta_sum_cmd(self, method='sumnummonien', tabname='T'):
        r"""
        Return a numerical approximation of the corresponding generalized multizeta
        """
        var_names = ['n%d'%n for n in range(1, self.length()+1)]

        # TODO: reorder the term to move out as much as possible the linear forms
        t = self._den._tuple

        if not t:
            return '1'

        cmd = "%s(%%s=1, %%s, %s)" %(method, tabname)
        d = len(t[0][0])
        for i in range(d-1):
            cmd = cmd % (var_names[i], '%s(%%s=1, %%s, %s)'% (method, tabname))

        tot = []
        for mon, mul in self._den:
            l = den.str_linear()
            if mul:
                l = '%s^%d' %(l, mul)
            tot.append(l)

        f = '1 / (%s)' % '*'.join(tot)
        cmd = cmd % (var_names[d-1], f)
        return cmd


# TODO: make it possible to use Laurent polynomials in denominator
# TODO: monomial substitution
# TODO: coefficient expansion Verdoolaege-Woods
class AbstractMSum(Element):
    r"""
    An abstract multiple sums for generating functions
    """
    def __init__(self, parent, data, allow_multiple=False, check=True):
        r"""
        _data is a dictionary: {denominator: polynomial}
        """
        Element.__init__(self, parent)
        V = self.parent().free_module()
        R = self.parent().laurent_polynomial_ring()
        self._data = {}

        if not data:
            return

        if isinstance(data, (tuple, list)):
            data = iter(data)

        elif isinstance(data, dict):
            if not check:
                self._data = data
                return
            data = list(data.items())

        elif isinstance(data, AbstractMSum):
            self._data = data.copy()
            return

        for term in data:
            if not isinstance(term, (tuple,list)) or len(term) != 2:
                raise ValueError
            den, num = term

            num = R(num)
            if num.is_zero():
                continue

            if not isinstance(den, FactoredDenominator):
                den = FactoredDenominator(den, V)

            if den in self._data:
                if not allow_multiple:
                    raise ValueError('multiple times the same denominator in the input')
                else:
                    self._data[den] += num
            else:
                self._data[den] = num

    def is_trivial_zero(self):
        r"""
        EXAMPLES::

            sage: from surface_dynamics.misc.multivariate_generating_series import MultivariateGeneratingSeriesRing

            sage: M = MultivariateGeneratingSeriesRing('x', 2)
            sage: M.term(0, []).is_trivial_zero()
            True
            sage: M.term(0, [((1,1),1)]).is_trivial_zero()
            True
            sage: M.term(1, []).is_trivial_zero()
            False
        """
        return not self._data

    def is_trivial_one(self):
        r"""
        EXAMPLES::

            sage: from surface_dynamics.misc.multivariate_generating_series import MultivariateGeneratingSeriesRing

            sage: M = MultivariateGeneratingSeriesRing('x', 2)
            sage: M.term(0, []).is_trivial_one()
            False
            sage: M.term(0, [((1,1),1)]).is_trivial_one()
            False
            sage: M.term(1, []).is_trivial_one()
            True
            sage: M.term(1, [((1,1),1)]).is_trivial_one()
            False

            sage: f = M.term(1, [((1,0),1)]) + M.term(1, [((0,1),1)])
            sage: f.is_trivial_one()
            False
        """
        if len(self._data) != 1:
            return False
        (den, num) = next(iteritems(self._data))
        return num.is_one() and den.is_one()

    def _add_(self, other):
        r"""
        TESTS::

            sage: from surface_dynamics.misc.multivariate_generating_series import MultivariateGeneratingSeriesRing

            sage: M = MultivariateGeneratingSeriesRing('x', 3)
            sage: m1 = M.term(1, [((1,1,0),1)])
            sage: m2 = M.term(-2, [((1,0,0),1),((0,1,0),2)])
            sage: m1 + m2
            (-2)/((1 - x1)^2*(1 - x0)) + (1)/((1 - x0*x1))
        """
        if self.is_trivial_zero():
            return other
        if other.is_trivial_zero():
            return self

        ans_data = self._data.copy()
        for den,num in other._data.items():
            if den in ans_data:
                ans_data[den] += num
                if ans_data[den].is_zero():
                    del ans_data[den]
            else:
                ans_data[den] = num

        R = self.parent()
        return R.element_class(R, ans_data)

    def _mul_(self, other):
        r"""
        TESTS::

            sage: from surface_dynamics.misc.multivariate_generating_series import MultivariateGeneratingSeriesRing

            sage: M = MultivariateGeneratingSeriesRing('x', 3)
            sage: m1 = M.term(1, [((1,1,0),1)])
            sage: m2 = M.term(-2, [((1,0,0),1),((0,1,0),2)])
            sage: m1 * m2
            (-2)/((1 - x1)^2*(1 - x0)*(1 - x0*x1))

            sage: R = M.laurent_polynomial_ring()
            sage: M.one() * R.one()
            (1)

            sage: R = M.laurent_polynomial_ring()
            sage: x0, x1, x2 = R.gens()
            sage: m1 = M.term(x0 - 2*x2 + 1, [((1,1,1),2)])
            sage: m2 = M.term(x0*x1 + 1, [((0,1,2),1)])
            sage: m3 = M.term(1 + x0 + x1, [((1,0,1),1)])
            sage: m4 = M.term(x0 - 1, [((1,0,0),2),((1,0,1),1)])
            sage: (m1 + m2) * (m3 + m4) - m1 * m3 - m2 * m3 - m1 * m4 - m2 * m4
            0
        """
        ans_data = {}
        for den1,num1 in self._data.items():
            for den2, num2 in other._data.items():
                den = den1 * den2
                num = num1 * num2
                if den in ans_data:
                    ans_data[den] += num
                    if not ans_data[den]:
                        del ans_data[den]
                else:
                    ans_data[den] = num

        M = self.parent()
        return M.element_class(M, ans_data)

    def __neg__(self):
        r"""
        TESTS::

            sage: from surface_dynamics.misc.multivariate_generating_series import MultivariateGeneratingSeriesRing

            sage: M = MultivariateGeneratingSeriesRing('x', 3)
            sage: m1 = M.term(1, [((1,1,0),1)])
            sage: m1
            (1)/((1 - x0*x1))
            sage: -m1
            (-1)/((1 - x0*x1))
            sage: m1 - m1
            0
        """
        ans_data = {}
        for den, num in self._data.items():
            ans_data[den] = -num
        M = self.parent()
        return M.element_class(M, ans_data)

    def _richcmp_(self, other, op):
        r"""
        TESTS::

            sage: from surface_dynamics.misc.multivariate_generating_series import MultivariateGeneratingSeriesRing

            sage: M = MultivariateGeneratingSeriesRing('x', 2)
            sage: x, y = M.laurent_polynomial_ring().gens()
            sage: M.term(x*y+1, [((1,1),1)]) == M.term(x*y+1, [((1,1),1)])
            True
        """
        if op != op_EQ and op != op_NE:
            raise TypeError('no comparison available for multivariate series')

        if self._data == other._data:
            return op == op_EQ

        (sden,snum), = self.factor()._data.items()
        (oden,onum), = other.factor()._data.items()

        if sden == oden and snum == onum:
            return op == op_EQ

        raise NotImplementedError

# TODO: here the elements are not necessarily with the same number of variables
# we want the following to work
# 1/3 * Sum(1/((n0)^4), n0=1..+oo) + 2/3 * Sum(1/((n0)*(n0 + n1)^3), n0=1..+oo, n1=1..+oo) + 1/3 * Sum(1/((n0)^2*(n0 + n1)^2), n0=1..+oo, n1=1..+oo)
class GeneralizedMultiZetaElement(AbstractMSum):
    def simplify(self):
        r"""
        Canonicalize each term by applying a permutation
        """
        from itertools import permutations

        P = self.parent()
        new_data = []
        n = P.ngens()
        for den,num in iteritems(self._data):
            # TODO: this is stupidly innefficient
            L = [list(k) + [v] for k,v in den._tuple]
            L.sort()
            Lmin = L
            for p in permutations(range(n)):
                LL = [[l[p[i]] for i in range(n)] + [l[n]] for l in L]
                LL.sort()
                if LL < Lmin:
                    Lmin = LL

            new_den = [(k[:-1], k[-1]) for k in Lmin]
            new_data.append((new_den, num))
        return P.element_class(P, new_data, allow_multiple=True)

    def expand(self):
        r"""
        Apply 1 / (L1 L2) = 1/(L1+L2) * (1/L1 + 1/L2) to try getting
        multiple zeta values
        """
        raise NotImplementedError

    def weights(self):
        r"""
        Return the set of weights of the generalized multiple zeta values.
        """
        return sorted(set(den.degree() for den in self._data))

    def _repr_(self):
        r"""
        String representation
        """
        terms = []
        var_names = ['n%d'%n for n in range(self.parent().laurent_polynomial_ring().ngens())]
        for den in sorted(self._data.keys()):
            num = self._data[den]
            fraction = str(num)
            if not den.is_one():
                fraction += ' * Sum(1/'
                linterm = den.str_linear(var_names)
                fraction += '(' + den.str_linear(var_names) + '), '
                fraction += ', '.join("%s=1..+oo"%v for v in var_names) + ')'
            terms.append(fraction)

        return ' + '.join(terms)

class MultivariateGeneratingSeries(AbstractMSum):
    def __nonzero__(self):
        return bool(self._data)

    def numerator(self):
        r"""
        Return the numerator if there is at most one summand.

        EXAMPLES::

            sage: from surface_dynamics.misc.multivariate_generating_series import MultivariateGeneratingSeriesRing
            sage: M = MultivariateGeneratingSeriesRing('x', 3)
            sage: x0, x1, x2 = M.laurent_polynomial_ring().gens()

            sage: f = M.term(x0 + x1^2*x2, [([1,3,0],1),([1,0,-1],1)])
            sage: f.numerator()
            x1^2*x2 + x0
            sage: g = M.term(1 - x2, [([1,1,0],1),([1,0,-1],2)])
            sage: (f + g).numerator()
            Traceback (most recent call last):
            ...
            ValueError: not a simple fraction
            sage: (f + g).factor().numerator()
            x0^2*x1^3 - x0*x1^3 + x0^3*x1*x2^-1 - x0^2*x1 - x0*x1^2 + x1^2*x2 - x0^2*x2^-1 + x0 - x2 + 1
        """
        if len(self._data) > 1:
            raise ValueError("not a simple fraction")
        if not self._data:
            return self.parent().laurent_polynomial_ring().zero()
        else:
            return list(self._data.values())[0]

    def summands(self):
        r"""
        Return the list of elementary summands

        EXAMPLES::

           sage: from surface_dynamics.misc.multivariate_generating_series import *
           sage: M = MultivariateGeneratingSeriesRing(2, 'x')
           sage: m1 = M.term(1, [((1,-1),1), ((0,1),1)])
           sage: m2 = M.term(1, [((-1,1),1), ((1,0),1)])
           sage: f = (m1 + m2)**2
           sage: for s in f.summands(): print(s)
           (2)/((1 - x0^-1*x1)*(1 - x1)*(1 - x0*x1^-1)*(1 - x0))
           (1)/((1 - x0^-1*x1)^2*(1 - x0)^2)
           (1)/((1 - x1)^2*(1 - x0*x1^-1)^2)
        """
        M = self.parent()
        res = []
        for den in sorted(self._data):
            num = self._data[den]
            res.append(M.term(num,den))
        return res

    # TODO: this is only working term by term. It completely ignores pole cancellation between
    # the terms
    #
    # 1 / (1 - x0 x1^-1) (1 - x1) + 1 / (1 - x0^-1 x1) (1 - x0)
    #
    #     sage: M = MultivariateGeneratingSeriesRing(2)
    #     sage: m1 = M.term(1, [((1,-1),1), ((0,1),1)])
    #     sage: m2 = M.term(1, [((-1,1),1), ((1,0),1)])
    #     sage: f = m1 + m2
    #
    # This is solved by applying the monomial substitution as in
    # Barvinok Woods 2003 "Short rational generating functions for lattice point problems"
    # (possible alternative in Verdoolaege)
    #
#    def monomial_substitution(self, arg, M=None):
#        r"""
#        Perform a monomial substitution
#
#        - ``arg`` - a list of monomials or vectors or a matrix
#
#        - ``M`` - an optional codomain
#
#        EXAMPLES::
#
#            sage: M = MultivariateGeneratingSeriesRing(2)
#            sage: R = M.laurent_polynomial_ring()
#            sage: x0,x1 = R.gens()
#
#            sage: f = M.term(x0, [((1,0),2),((0,1),1)])
#            sage: f
#            (x0)/((1 - x1)*(1 - x0)^2)
#            sage: f.monomial_substitution(matrix(2, [1,1,1,1]))
#            (x0*x1)/((1 - x0*x1)^3)
#
#            sage: C = MultivariateGeneratingSeriesRing(3, 'z')
#            sage: f.monomial_substitution(matrix(2, 3, [1,0,1,0,1,1]), C)
#            (z0*z2)/((1 - z1*z2)*(1 - z0*z2)^2)
#        """
#        M_domain = self.parent()
#        R_domain = M_domain.laurent_polynomial_ring()
#
#        if M is None:
#            M_codomain = M_domain
#            R_codomain = R_domain
#        else:
#            M_codomain = M
#            R_codomain = M.laurent_polynomial_ring()
#
#        if arg.nrows() != M_domain.ngens() or arg.ncols() != M_codomain.ngens():
#            raise ValueError('wrong number of rows or columns')
#
#        s_dict = {R_domain.gen(i): R_codomain.monomial(*row) for i,row in enumerate(arg)}
#
#        # the monomial substitution also depends on the numerator !!
#        new_data = {}
#        for den, num in self._data.items():
#            monomial_substitution(arg, num, den, M_codomain)
#            assert num.parent() is R_domain
#            num = R_codomain(num.subs(s_dict))
#            if den in new_data:
#                new_data[den] += num
#            else:
#                new_data[den] = num
#
#        return M_codomain.element_class(M_codomain, new_data)

    # TODO: this does not make any sense
    # the residue is 1 / L1(h) ... Ld(h) where the Li are linear forms. At
    # no point we are asking for a sum of them!!
    # In other words, this method should just return a function h -> 1 / prod(Li)
    def residue(self):
        r"""
        denominator: each (1 - mon)^k in denom is replaced with -> mon^k
        numerator: evaluate at (1,1,...,1)

        OUTPUT: a pair '(degree, value)`

        EXAMPLES::

            sage: from surface_dynamics.misc.multivariate_generating_series import MultivariateGeneratingSeriesRing

            sage: M = MultivariateGeneratingSeriesRing(2, 'x')
            sage: R = M.laurent_polynomial_ring()
            sage: x0,x1 = R.gens()
            sage: f = M.term(x0, [((1,1),2)])
            sage: f.residue()
            (2, [(1, {(1, 1): 2})])
            sage: f = M.term(x0, [((1,1),2)]) + M.term(1, [((1,0),1),((0,1),1),((1,1),1)])
            sage: r = f.residue()
            sage: r # random
            (3, [(1, {(1, 0): 1, (0, 1): 1, (1, 1): 1})])
            sage: f = M.term(x0, [((1,1),2)]) + M.term(1, [((1,0),1),((1,1),1)])
            sage: r = f.residue()
            sage: r # random
            (2, [(1, {(1, 1): 2}), (1, {(1, 0): 1, (1, 1): 1})])
        """
        R = self.parent().laurent_polynomial_ring()
        one = QQ.one()
        values = {g: one for g in R.gens()}
        ans = []
        d = -1
        for den, num in self._data.items():
            if den.degree() >= d:
                if den.degree() > d:
                    ans = []
                    d = den.degree()
                num = QQ(num.subs(values))
                if num.is_zero():
                    raise NotImplementedError('zero numerator')
                ans.append((num, den))
        return d, ans

    def derivative(self, var):
        r"""
        EXAMPLES::

            sage: from surface_dynamics.misc.multivariate_generating_series import MultivariateGeneratingSeriesRing

            sage: M = MultivariateGeneratingSeriesRing('x', 2)
            sage: R = M.laurent_polynomial_ring()
            sage: x0, x1 = R.gens()

            sage: f = M.term(1, [((1,0),1)])
            sage: f.derivative(0)
            (1)/((1 - x0)^2)
            sage: f.derivative(0).derivative(0)
            (2)/((1 - x0)^3)

            sage: xx0, xx1 = R.polynomial_ring().gens()
            sage: f.taylor(10).derivative(xx0) == f.derivative(0).taylor(9)
            True
            sage: f.derivative(1)
            0

            sage: f = M.term(x0 * x1, [((1,1),2)])
            sage: f.derivative(0)
            (x1)/((1 - x0*x1)^2) + (2*x0*x1^2)/((1 - x0*x1)^3)
            sage: f.derivative(0).taylor(10) - f.taylor(10).derivative(xx0)
            30*x0^5*x1^6

            sage: f = M.term(1, [((1,2),2), ((0,1),2), ((1,0),2), ((1,1),2)])
            sage: min((f.taylor(10).derivative(xx0) - f.derivative(0).taylor(10)).degrees())
            9
            sage: min((f.taylor(10).derivative(xx1) - f.derivative(1).taylor(10)).degrees())
            10

            sage: f = M.term(1, [((1,0),2), ((1,1),1)])
            sage: f.taylor(10).derivative(xx0).derivative(xx1) - f.derivative(0).derivative(1).taylor(9)
            -18*x0^9*x1 - 42*x0^8*x1^2 - ... - 72*x0^5*x1^3 - 25*x0^4*x1^4

            sage: f = M.zero() + M.term(1/1*x0^2 * x1, [((1,0),1), ((1,1),1)])
            sage: f.derivative(0)
            (2*x0*x1)/((1 - x0)*(1 - x0*x1)) + (x0^2*x1^2)/((1 - x0)*(1 - x0*x1)^2) + (x0^2*x1)/((1 - x0)^2*(1 - x0*x1))

        You can indistinctly use integers, strings or polynomial variables for ``var``::

            sage: f = M.term(1, [((1,1),1)])
            sage: f.derivative('x0')
            (x1)/((1 - x0*x1)^2)
            sage: f.derivative(0)
            (x1)/((1 - x0*x1)^2)
            sage: f.derivative(R.gen(0))
            (x1)/((1 - x0*x1)^2)

        Checking errors in the input::

            sage: f.derivative(-1)
            Traceback (most recent call last):
            ...
            ValueError: generator not defined
            sage: f.derivative('q')
            Traceback (most recent call last):
            ...
            ValueError: 'q' not valid as a variable
        """
        M = self.parent()
        R = M.laurent_polynomial_ring()

        if isinstance(var, numbers.Integral):
            j = var
        else:
            try:
                j = R.gens().index(R(var))
            except (ValueError, TypeError):
                raise ValueError('%r not valid as a variable' % var)

        var = R.gen(j)

        ans = M.zero()
        for den, num in self._data.items():
            ans += M.term(num.derivative(var), den)
            s = sum(mul * laurent_monomial(R, v) * M.term(1, mon) for mul, v, mon in den.logarithmic_minus_derivative(j))
            ans += M.term(num, den) * s
        return ans

    def derivative_up_to_lower_order_terms(self, var):
        r"""
        Each term ``u/v`` is replaced by ``-u v'/v^2``.

        This corresponds to one half of the derivative, the other half
        being ``u' / v``. This second part can be ignored when asymptotics
        question are considered.

        EXAMPLES::

            sage: from surface_dynamics.misc.multivariate_generating_series import MultivariateGeneratingSeriesRing

            sage: M = MultivariateGeneratingSeriesRing('x', 2)
            sage: f = M.term(1, [((1,0),2)])
            sage: f
            (1)/((1 - x0)^2)
            sage: f.derivative_up_to_lower_order_terms(0)
            (2)/((1 - x0)^3)

            sage: f = M.term(1, [((1,0),1),((0,1),1),((1,1),1)])
        """
        M = self.parent()
        R = M.laurent_polynomial_ring()

        ans = M.zero()
        for den, num in self._data.items():
            s = M.zero()
            for m,v,d in den.logarithmic_minus_derivative(var):
                s += M.term(m * laurent_monomial(R,v), [(d,1)])
            ans += M.term(num, den) * s

        return ans

    def delta(self):
        r"""
        Take a derivative (up to lower order terms) with respect to each of the variables.
        """
        a = self
        for i in range(self.parent().ngens()):
            a = a.derivative_up_to_lower_order_terms(i)
        return a

    # what does prec means for Laurent polynomials!?
    def taylor(self, prec, R=None):
        r"""
        EXAMPLES::

            sage: from surface_dynamics.misc.multivariate_generating_series import MultivariateGeneratingSeriesRing

            sage: M = MultivariateGeneratingSeriesRing('t,u')
            sage: t,u = M.laurent_polynomial_ring().gens()
            sage: f = M.term(t + u^3, [((1,1), 1), ((1,2),2)])
            sage: f
            (u^3 + t)/((1 - t*u)*(1 - t*u^2)^2)
            sage: f.taylor(10)
            2*t^4*u^8 + 4*t^3*u^9 + t^4*u^7 + 3*t^3*u^8 + ... + t*u^4 + 2*t^2*u^2 + t^2*u + u^3 + t

        TODO: this is only working term by term but for example containing negative
        powers as::

            sage: M = MultivariateGeneratingSeriesRing('x', 2)   # not tested
            sage: m1 = M.term(1, [((1,-1),1), ((0,1),1)])   # not tested
            sage: m2 = M.term(1, [((-1,1),1), ((1,0),1)])   # not tested
            sage: f = m1 + m2                               # not tested
            sage: f.taylor(10)                              # not tested

        TESTS::

            sage: M = MultivariateGeneratingSeriesRing('x', 2)
            sage: R = M.laurent_polynomial_ring()
            sage: x0, x1 = R.gens()
            sage: type(M.term(1, [((1,0),1)]).taylor(10))
            <type 'sage.rings.polynomial.multi_polynomial_libsingular.MPolynomial_libsingular'>
        """
        M = self.parent()
        R = M.laurent_polynomial_ring().polynomial_ring()
        S = M.laurent_polynomial_ring().polynomial_ring()['EXTRA_VAR']
        ans = S.zero()
        for den, num in self._data.items():
            ans += (S(num) * den.inverse_series_trunc(S, prec)).truncate(prec)
        return R(ans.subs({S.gen(): R.one()}))

    def is_zero(self):
        if self.is_trivial_zero():
            return True
        raise NotImplementedError

    def is_one(self):
        if self.is_trivial_one():
            return True
        raise NotImplementedError

    def degrees(self):
        return (den.degree() for den in self._data.keys())

    def _repr_(self):
        r"""
        TESTS::

            sage: from surface_dynamics.misc.multivariate_generating_series import MultivariateGeneratingSeriesRing

            sage: M = MultivariateGeneratingSeriesRing('x', 2)
            sage: R = M.laurent_polynomial_ring()
            sage: x0, x1 = R.gens()
            sage: M.term(-x0 + x1^3, [((1,1), 1), ((1,2),2)])
            (x1^3 - x0)/((1 - x0*x1)*(1 - x0*x1^2)^2)

            sage: M = MultivariateGeneratingSeriesRing('t,u')
            sage: t, u = M.laurent_polynomial_ring().gens()
            sage: M.term(-t + u^3, [((1,1), 1), ((1,2),2)])
            (u^3 - t)/((1 - t*u)*(1 - t*u^2)^2)
        """
        if not self._data:
            return '0'

        var_names = self.parent().laurent_polynomial_ring().variable_names()
        terms = []
        for den in sorted(self._data):
            num = self._data[den]
            fraction = '(' + str(num) + ')'
            if not den.is_one():
                fraction += '/'
                fraction += '(' + den.str_monomials(var_names) + ')'

            terms.append(fraction)

        return ' + '.join(terms)

    def subs_numerator(self, *args, **kwds):
        r"""
        TESTS::

            sage: from surface_dynamics.misc.multivariate_generating_series import MultivariateGeneratingSeriesRing
            sage: M = MultivariateGeneratingSeriesRing('x', 2)
            sage: R = M.laurent_polynomial_ring()
            sage: x0, x1 = R.gens()
            sage: f = M.term(-x0 + 2*x0*x1^3, [((1,1), 1), ((1,2),2)])
            sage: f.subs_numerator(x0=1, x1=1)
            (1)/((1 - x0*x1)*(1 - x0*x1^2)^2)
        """
        M = self.parent()
        res = M.zero()
        for den, num in self._data.items():
            res += M.term(num.subs(*args, **kwds), den)
        return res

    def factor(self, simplify=False):
        """
        Group all the partial fractions into a unique fraction.

        EXAMPLES::

            sage: from surface_dynamics.misc.multivariate_generating_series import MultivariateGeneratingSeriesRing
            sage: M = MultivariateGeneratingSeriesRing('x', 3)

            sage: f = M.term(1, [([1,3,0],1),([1,0,-1],1)])
            sage: g = M.term(1, [([1,1,0],1),([1,0,-1],2)])
            sage: (f + g).factor()
            (-x0*x1^3 + x0^2*x1*x2^-1 - x0*x1 - x0*x2^-1 + 2)/((1 - x0*x2^-1)^2*(1 - x0*x1)*(1 - x0*x1^3))

            sage: x0, x1, x2 = M.laurent_polynomial_ring().gens()
            sage: f = M.term(x0 + x1, [([1,3,0],1),([1,0,1],1)])
            sage: g = M.term(x2 - 1, [([1,1,0],1),([1,0,1],2)])
            sage: h = f + g
            sage: t1 = h.taylor(8)
            sage: t2 = h.factor().taylor(8)
            sage: assert all(sum(e) >= 8 for e in (t1 - t2).exponents())

        Simplification::

            sage: f1 = M.term(x0, [([1,0,0],1)])
            sage: f2 = M.term(x1, [([0,1,0],1)])
            sage: f3 = M.term(x0*x1 - 1, [([1,0,0],1), ([0,1,0],1)])
            sage: (f1 + f2 + f3).factor()
            (-1)
        """
        M = self.parent()

        # compute the product of denominators
        V = M.free_module()
        D = FactoredDenominator([], V)
        for den, num in self._data.items():
            D.lcm_update(den)

        L = M.laurent_polynomial_ring()
        N = L.zero()
        for den,num in self._data.items():
            N += num * (D / den).to_polynomial(L)

        # Look for simplifications. In the case of Laurent polynomials, we do not
        # do anything.
        # TODO: do something for Laurent polynomials
        if any(i < 0 for den in self._data for v in den._dict for i in v) or \
           any(i < 0 for num in self._data.values() for e in num.exponents() for i in e):
            return M.term(N, D)

        P = L.polynomial_ring()
        N = P(N)
        gt = False
        for mon,mult in list(D._dict.items()):
            pmon = P.one() - P.monomial(*mon)
            q,r = N.quo_rem(pmon)
            while mult and not r:
                gt = True
                mult -= 1
                N = q
                q,r = N.quo_rem(pmon)

            if mult:
                D._dict[mon] = mult
            else:
                del D._dict[mon]

        if gt:
            D._tuple = tuple(sorted(D._dict.items()))

        return M.term(N, D)

    def as_symbolic(self):
        from sage.symbolic.ring import SR
        return SR(str(self))


class AbstractMSumRing(Parent):
    def __init__(self, *args):
        if len(args) == 1 and isinstance(args[0], AbstractMSumRing):
            self._laurent_polynomial_ring = args[0]._laurent_polynomial_ring
            self._free_module = args[0]._free_module
            self._laurent_polynomial_ring_extra_var = args[0]._laurent_polynomial_ring_extra_var
        else:
            from sage.rings.polynomial.laurent_polynomial_ring import LaurentPolynomialRing
            from sage.modules.free_module import FreeModule
            self._laurent_polynomial_ring = LaurentPolynomialRing(QQ, *args)
            dim = ZZ(self._laurent_polynomial_ring.ngens())
            self._free_module = FreeModule(ZZ, dim)

            # univariate extension of the polynomial ring
            # (needed in several algorithms)
            self._laurent_polynomial_ring_extra_var = self._laurent_polynomial_ring['EXTRA_VAR']

        Parent.__init__(self, category=Rings())

    def ngens(self):
        return self.laurent_polynomial_ring().ngens()

    def free_module(self):
        return self._free_module

    def polynomial_ring(self):
        raise ValueError

    def polynomial_ring_extra_var(self):
        raise ValueError

    def laurent_polynomial_ring(self):
        return self._laurent_polynomial_ring

    def laurent_polynomial_ring_extra_var(self):
        return self._laurent_polynomial_ring_extra_var

    def with_extra_var(self):
        return self['EXTRA_VAR']

    @cached_method
    def zero(self):
        r"""
        EXAMPLES::

            sage: from surface_dynamics.misc.multivariate_generating_series import MultivariateGeneratingSeriesRing

            sage: M = MultivariateGeneratingSeriesRing('x', 2)
            sage: M.zero()
            0
            sage: M.zero().parent() is M
            True
            sage: M.zero().is_zero()
            True
        """
        return self._element_constructor_(QQ.zero())

    @cached_method
    def one(self):
        r"""
        EXAMPLES::

            sage: from surface_dynamics.misc.multivariate_generating_series import MultivariateGeneratingSeriesRing

            sage: M = MultivariateGeneratingSeriesRing('x', 2)
            sage: M.zero()
            0
            sage: M.zero().parent() is M
            True
            sage: M.one().is_one()
            True
        """
        return self._element_constructor_(QQ.one())

    def term(self, num, den):
        r"""
        Return the term ``num / den``.

        INPUT:

        - ``num`` - a Laurent polynomial

        - ``den`` - a list of pairs ``(vector, power)`` or a dictionary
           whose keys are the vectors and the values the powers. The
           vector ``v = (v_0, v_1, \ldots)`` with power ``n`` corresponds
           to the factor `(1 - x_0^{v_0} x_1^{v_1} \ldots x_k^{v_k})^n`.

        EXAMPLES::

            sage: from surface_dynamics.misc.multivariate_generating_series import MultivariateGeneratingSeriesRing

            sage: M = MultivariateGeneratingSeriesRing('x', 3)
            sage: M.term(1, [([1,1,0],1),([1,0,-1],2)])
            (1)/((1 - x0*x2^-1)^2*(1 - x0*x1))
            sage: M.term(1, {(1,1,0): 1, (1,0,2): 2})
            (1)/((1 - x0*x2^2)^2*(1 - x0*x1))

        Also works if the denominator is already a factored denominator::

            sage: from surface_dynamics.misc.factored_denominator import FactoredDenominator
            sage: M = MultivariateGeneratingSeriesRing('x', 3)
            sage: den = FactoredDenominator({(1,0,0): 2, (1,1,1):1})
            sage: M.term(1, den)
            (1)/((1 - x0)^2*(1 - x0*x1*x2))

        TESTS::


            sage: from surface_dynamics.misc.multivariate_generating_series import MultivariateGeneratingSeriesRing
            sage: M = MultivariateGeneratingSeriesRing('x', 2)
            sage: R = M.laurent_polynomial_ring()
            sage: x0, x1 = R.gens()
            sage: d = {}
            sage: v0 = vector(QQ, (1,0)); v0.set_immutable()
            sage: v1 = vector(QQ, (1,1)); v1.set_immutable()
            sage: d[v1] = 1
            sage: d[v0] = 1
            sage: M.term(1*x0^2*x1, d)
            (x0^2*x1)/((1 - x0)*(1 - x0*x1))
        """
        return self.element_class(self, [(den, num)], self.free_module())

    def _element_constructor_(self, arg):
        r"""
        TESTS::

            sage: from surface_dynamics.misc.multivariate_generating_series import MultivariateGeneratingSeriesRing

            sage: M = MultivariateGeneratingSeriesRing('x', 2)
            sage: M(1)
            (1)

            sage: R = M.laurent_polynomial_ring()
            sage: M(R.0)
            (x0)
        """
        num = self._laurent_polynomial_ring(arg)
        return self.element_class(self, [([], num)], self.free_module())

class GeneralizedMultiZetaRing(AbstractMSumRing):
    Element = GeneralizedMultiZetaElement

    def _coerce_map_from_(self, other):
        if self.laurent_polynomial_ring().base_ring().has_coerce_map_from(other):
            return True


# TODO: this should actually be an algebra over QQ[x0, x1, ..., xn]
# TODO: we should have two versions, as an algebra over the polynomial
#       ring QQ[x0, x1, ..., xn] and as an algebra over the Laurent
#       polynomial ring QQ[x0, x0^-1, x1, x1^-1, ..., xn, xn^-1]
class MultivariateGeneratingSeriesRing(AbstractMSumRing):
    r"""
    EXAMPLES::

        sage: from surface_dynamics.misc.multivariate_generating_series import MultivariateGeneratingSeriesRing

        sage: MultivariateGeneratingSeriesRing('x', 3)
        Multivariate quasi-polynomial generating series on x0, x1, x2

        sage: MultivariateGeneratingSeriesRing(['x', 'y', 'z'])
        Multivariate quasi-polynomial generating series on x, y, z

        sage: MultivariateGeneratingSeriesRing('a,b,c')
        Multivariate quasi-polynomial generating series on a, b, c

        sage: MultivariateGeneratingSeriesRing('y', 2)
        Multivariate quasi-polynomial generating series on y0, y1

        sage: M = MultivariateGeneratingSeriesRing('x', 3)
        sage: M.zero()
        0
        sage: M.one()
        (1)
        sage: m1 = M.term(1, [((1,1,0),1)])
        sage: m1
        (1)/((1 - x0*x1))
        sage: m2 = M.term(-2, [((1,0,0),1),((0,1,0),2)])
        sage: m2
        (-2)/((1 - x1)^2*(1 - x0))
    """
    Element = MultivariateGeneratingSeries

    @cached_method
    def residue_ring(self):
        return GeneralizedMultiZetaRing(self)

    def _repr_(self):
        vars_string = ', '.join(self.laurent_polynomial_ring().variable_names())
        return 'Multivariate quasi-polynomial generating series on ' + vars_string

    def _coerce_map_from_(self, other):
        r"""
        EXAMPLES::

            sage: from surface_dynamics.misc.multivariate_generating_series import MultivariateGeneratingSeriesRing

            sage: M = MultivariateGeneratingSeriesRing('x', 2)

            sage: M.has_coerce_map_from(ZZ)
            True
            sage: M.coerce_map_from(ZZ)
            Co...ion map:
              From: Integer Ring
              To:   Multivariate quasi-polynomial generating series on x0, x1

            sage: M.has_coerce_map_from(QQ)
            True

            sage: M.has_coerce_map_from(M.laurent_polynomial_ring())
            True
        """
        if self.laurent_polynomial_ring().has_coerce_map_from(other):
            return True



# ****************************************************************************
#       Copyright (C) 2011-2019 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from .origami_dense import PillowcaseCover_dense_pyx

try:
    # Trac #28652: Rework the constructor of PermutationGroupElement
    from sage.groups.perm_gps.constructor import PermutationGroupElement
except ImportError:
    from sage.groups.perm_gps.permgroup import PermutationGroupElement


def PillowcaseCover(g0, g1, g2, g3=None,
        sparse=False,
        check=True,
        as_tuple=False,
        positions=None, name=None):
    r"""
    Pillowcase cover constructor.

    The chosen flat structure is as follows

       3-----2-----3
       |     .     |
       |     .     |
       |     .     |
       0-----1-----0
    """
    if not as_tuple:
        g0 = PermutationGroupElement(g0, check=check)
        g1 = PermutationGroupElement(g1, check=check)
        g2 = PermutationGroupElement(g2, check=check)
        if g3 is None:
            g3 = (~g2) * (~g1) * (~g0)
        else:
            g3 = PermutationGroupElement(g3, check=check)

        g0 = [i-1 for i in g0.domain()]
        g1 = [i-1 for i in g1.domain()]
        g2 = [i-1 for i in g2.domain()]
        g3 = [i-1 for i in g3.domain()]

        N = max([len(g0),len(g1),len(g2),len(g3)])
        g0.extend(range(len(g0),N))
        g1.extend(range(len(g1),N))
        g2.extend(range(len(g2),N))
        g3.extend(range(len(g3),N))

    elif check:
        s0 = set(g0)
        s1 = set(g1)
        s2 = set(g2)
        s3 = set(g3)
        N = len(g0)
        if len(g1) != N or len(g2) != N or len(g3) != N:
            raise ValueError("the four tuples must be of the same length")
        for i in range(N):
            if not i in g0:
                raise ValueError("%d is not in g0=%s" %(i,str(g0)))
            if not i in g1:
                raise ValueError("%d is not in g1=%s" %(i,str(g1)))
            if not i in g2:
                raise ValueError("%d is not in g2=%s"%(i,str(g2)))
            if not i in g3:
                raise ValueError("%d is not in g3=%s"%(i,str(g3)))

    pcc = PillowcaseCover_dense(tuple(g0),tuple(g1),tuple(g2),tuple(g3))

    if name is not None:
        pcc.rename(name)
#    if positions is not None:
#        o.set_positions(positions)
    if check:
        pcc._check()
    return pcc

class PillowcaseCover_dense(PillowcaseCover_dense_pyx):
    r"""
    Generic class for pillowcase cover.
    """
    def __repr__(self):
        r"""
        TESTS::

            sage: from surface_dynamics import *
            sage: PillowcaseCover('(1,2)(3,4)', '(1,3)', '()')   # indirect doctest
            g0 = (1,2)(3,4)
            g1 = (1,3)(2)(4)
            g2 = (1)(2)(3)(4)
            g3 = (1,4,3,2)
        """
        return '\n'.join("g%d = %s"%(i,self.g(i).cycle_string(True)) for i in range(4))

    def g(self, i=None):
        r"""
        Return the ``i``-th permutation that defines this pillowcase cover.
        """
        if i is None:
            return self.g(0), self.g(1), self.g(2), self.g(3)

        i = int(i)
        if i < 0 or i > 3:
            raise IndexError("the index i (={}) must be in {{0,1,2,3}}".format(i))

        return PermutationGroupElement([j+1 for j in self.g_tuple(i)], check=False)

    def _check(self):
        x = self.g(0) * self.g(1) * self.g(2) * self.g(3)
        if not x.is_one():
            raise ValueError

    def monodromy(self):
        r"""
        Return the monodromy group of the pillowcase cover.

        The monodromy group of an origami is the group generated by the
        permutations `g_i` for `i` in 0,1,2,3.
        """
        from sage.groups.perm_gps.permgroup import PermutationGroup
        return PermutationGroup(self.g())

    def as_graph(self):
        r"""
        Return the graph associated to self
        """
        from sage.graphs.digraph import DiGraph

        G = DiGraph(multiedges=True,loops=True)
        d = self.degree()
        g = [self.g_tuple(i) for i in range(4)]
        for i in range(d):
            for j in range(4):
                G.add_edge(i,g[j][i],j)
        return G

    def is_connected(self):
        r"""
        Check whether the origami is connected or not

        It is equivalent to ask whether the group generated by `r` and `u` acts
        transitively on the `\{1,\dots,n\}`.
        """
        return self.as_graph().is_connected()

    def connected_components(self):
        r"""
        Return the list of connected origami that composes this origami.
        """
        cc = self.as_graph().connected_components()
        g = [self.g_tuple(i) for i in range(4)]
        if len(cc) == 1:
            return [self]
        l = []
        for c in cc:
            gg = [[None] * len(c) for _ in range(4)]
            d = dict((c[i],i) for i in range(len(c)))
            for i in c:
                for j in range(4):
                    gg[j][d[i]] = d[g[j][i]]
            l.append(Pillowcase_cover(g[0],g[1],g[2],g[3],check=True,as_tuple=True))
        return l

    def is_orientable(self):
        r"""
        Test whether the foliation is orientable.
        """
        return self.as_graph().to_undirected().is_bipartite()

    def profile(self,i=None):
        r"""
        Return the profile (= ramification type above each pole).
        """
        if i is None:
            return [self.profile(i) for i in range(4)]
        return sorted((len(c) for c in self.g(i).cycle_tuples(singletons=True)),reverse=True)

    def stratum(self,fake_zeros=False):
        r"""
        Return the stratum of self. It may be either a stratum of Abelian or
        quadratic differentials.

        EXAMPLES::

            sage: from surface_dynamics import *
            sage: PillowcaseCover('(1,2)(3,4)', '(1,3)', '()').stratum()
            Q_0(2, -1^6)
        """
        p = sum(self.profile(),[])
        if self.is_orientable():
            from surface_dynamics.flat_surfaces.abelian_strata import AbelianStratum
            if fake_zeros:
                zeros = [(i-2)//2 for i in p]
            else:
                zeros = [(i-2)//2 for i in p if i != 2]
            if zeros == []:
                return AbelianStratum([0])
            return AbelianStratum(zeros)

        else:
            from surface_dynamics.flat_surfaces.quadratic_strata import QuadraticStratum
            if fake_zeros:
                zeros = [i-2 for i in p]
            else:
                zeros = [i-2 for i in p if i != 2]
            return QuadraticStratum(zeros)

    def is_primitive(self, return_base=False):
        r"""
        A pillowcase cover is primitive if it does not cover an other pillowcase
        cover.
        """
        from sage.arith.all import is_prime
        if is_prime(self.degree()):
            return True

        return bool(gap.IsPrimitive(self.monodromy()))

    def orientation_cover(self):
        r"""
        Return the orientation cover as an origami.

        EXAMPLES:

        The pillowcase itself has cover a torus (made from 4 squares)::

            sage: from surface_dynamics import *

            sage: p0 = p1 = p2 = p3 = [0]
            sage: pc = PillowcaseCover(p0, p1, p2, p3, as_tuple=True)
            sage: pc.stratum()
            Q_0(-1^4)
            sage: o = pc.orientation_cover()
            sage: o
            (1,2)(3,4)
            (1,4)(2,3)
            sage: o.stratum()
            H_1(0)

        An example in Q(1,-1^5) whose cover belongs to H(2)::

            sage: p0 = [2,1,0]
            sage: p1 = [2,0,1]
            sage: p2 = [1,0,2]
            sage: p3 = [0,1,2]
            sage: pc = PillowcaseCover(p0, p1, p2, p3,as_tuple=True)
            sage: pc.stratum()
            Q_0(1, -1^5)
            sage: o = pc.orientation_cover()
            sage: o
            (1,2,3,4)(5,6)(7,10,9,8)(11,12)
            (1,10,5,12)(2,9)(3,8)(4,7,6,11)
            sage: o.stratum()
            H_2(2)

        A last example in Q(2^2)::

            sage: q = QuadraticCylinderDiagram('(0,1)-(2,3) (0,3)-(1,2)')
            sage: pc = q.cylcoord_to_pillowcase_cover([1,1,1,1], [2,2], [0,1])
            sage: pc.orientation_cover().stratum()
            H_3(1^4)
        """
        from surface_dynamics.misc.permutation import perm_invert
        from surface_dynamics.flat_surfaces.origamis.origami import Origami

        g0 = self.g_tuple(0)
        g1 = self.g_tuple(1)
        g2 = self.g_tuple(2)
        g3 = self.g_tuple(3)

        n = len(g0)

        r = [None] * (4*n)
        u = [None] * (4*n)

        h1 = perm_invert(g1)
        h2 = perm_invert(g2)

        for i in range(n):
            r[2*i] = 2*i+1
            r[2*i+1] = 2*g3[g2[i]]
            r[2*n+2*i+1] = 2*n + 2*i
            r[2*n+2*i] = 2*n + 2*g1[g0[i]] + 1

            u[2*i] = 2*n + 2*h2[i] + 1
            u[2*i+1] = 2*n + 2*g2[i]
            u[2*n+2*i] = 2*g1[i] + 1
            u[2*n+2*i+1] = 2*h1[i]

        return Origami(r, u, as_tuple=True)
